/* Copyright (C) 2004-2013 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana - C3SL/UFPR
 *
 * This file is part of datasid
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 * USA.
 */

#include "client.h"

/* Summary: Get parameters passed by user and contact webservice server
 *              to verify client version and send availability information.
 * Usage: client <url> <inep> <mac> [proxy_host [proxy_port [proxy_user
 *                                                           proxy_passwd]]]
 * Return: Exit status:
 *          1) Parameter error
 *          2) Soap error
 *          3) Version file error (open, memory allocation)
 *          4) Client out-of-date
 *          5) Server returned an error
 *          6) (Not used anymore)
 *          7) Server returned an empty version string
 *          8) Server returned an invalid version string
 *          9) Server did not find the version file
 *         10) Server returned an empty update link
 *         11) Invalid INEP error
 *         12) XML file error (open, memory allocation)
 */

/*----------------------------------------------------------------------------
 * Summary: Print to standard error the help messages.
 * Parameters: void
 * Return: void
 */
void print_help(void)
{
    ERROR("USAGE: datasid-gsoap-client [OPTIONS]");
    ERROR("\n\n");
    ERROR("-i, --send-inventory=XML   Send the collected xml ");
    ERROR("-N, --send-net-usage=XML   Send the collected xml ");
    ERROR("-u, --update=VERSION  Check if VERSION is the latest");
    ERROR("                      (and print the update link)");
    ERROR("-l, --url=URL         Web Service URL");
    ERROR("-H, --phost=host      Proxy host\n");
    ERROR("-p, --pport=port      Proxy port\n");
    ERROR("-U, --puid=userid     Proxy user id\n");
    ERROR("-w, --ppasswd=pswd    Proxy password\n");
    ERROR("-h, --help            Show this help\n");
    ERROR("-n, --nic=eth0        Network interface to be bound\n");
}

int main(int argc, char **argv)
{
    extern char *optarg;
    extern int optind, opterr, optopt;
    int option, option_index;
    int do_send_inventory, do_update, do_send_net_usage;

    char *url, *xml, *version, *device;

    char *proxy_host, *proxy_userid, *proxy_passwd;
    int proxy_port;

    struct soap *soap;

    static struct option long_options[] =
    {
        {"help",        no_argument, 0, 'h'},
        {"send-inventory",   required_argument, 0, 'i'},
        {"send-net-usage",   required_argument, 0, 'N'},
        {"update",      required_argument, 0, 'u'},
        {"url",         required_argument, 0, 'l'},
        {"phost",       required_argument, 0, 'H'},
        {"pport",       required_argument, 0, 'p'},
        {"puid",        required_argument, 0, 'U'},
        {"ppasswd",     required_argument, 0, 'w'},
        {"nic",         required_argument, 0, 'n'},
        {0, 0, 0, 0}
    };

    do_send_net_usage = do_send_inventory = do_update = 0;
    url = xml = version = device = NULL;

    proxy_host = proxy_userid = proxy_passwd = NULL;
    proxy_port = 0;

    /* Parse command line */
    while ((option = getopt_long(argc, argv, "hNi:u:l:H:p:U:w:n:",
                                 long_options, &option_index)) != -1)
    {
        switch (option)
        {
            case 'i':
                xml = optarg;
                do_send_inventory = 1;
                break;
            case 'N':
                xml = optarg;
                do_send_net_usage = 1;
                break;
            case 'u':
                version = optarg;
                do_update = 1;
                break;
            case 'l':
                url = optarg;
                break;
            case 'H':
                proxy_host = optarg;
                break;
            case 'p':
                proxy_port = atoi(optarg);
                break;
            case 'U':
                proxy_userid = optarg;
                break;
            case 'w':
                proxy_passwd = optarg;
                break;
            case 'h':
                print_help();
                exit(0);
                break;
            case 'n':
                device = optarg;
                break;
            case '?':
                ERROR("ERROR (1): ");
                print_help();
                exit(1);
                break;
            default:
                break;
        }
    }

    /* check if url is set */
    if (url == NULL)
    {
        ERROR("ERROR (1): Missing URL parameter.\n");
        print_help();
        exit(1);
    }

    /* check if none or more than one actions was choosen */
    if (do_send_net_usage + do_send_inventory + do_update != 1)
    {
        ERROR("ERROR (1): Use one, and only one of the following options:\n-i, --send-inventory\n-N, --send-net-usage\n-u, --update\n");
        print_help();
        exit(1);
    }

    if (optind < argc)
    {
        ERROR("ERROR (1): Unknown option(s): ");
        while (optind < argc - 1)
        {
            ERROR("%s, ", argv[optind++]);
        }
        ERROR("%s\n", argv[optind++]);
        print_help();
        exit(1);
    }

    /* initialize soap */
    soap = soap_new();

    /* set device */
    if (device != NULL)
    {
        soap->device = malloc(sizeof(char) * strlen(device));
        strcpy(soap->device, device);
        DEBUG("Network interface device \"%s\" used\n", soap->device);
    }

    /* set proxy */
    if (proxy_host != NULL)
        soap->proxy_host = proxy_host;

    if (proxy_userid != NULL)
        soap->proxy_userid = proxy_userid;

    if (proxy_passwd != NULL)
        soap->proxy_passwd = proxy_passwd;

    if (proxy_port != 0)
        soap->proxy_port = proxy_port;

    /* set timeout */
    soap->send_timeout = TIMEOUT;
    soap->recv_timeout = TIMEOUT;

    if (do_update == 1)
        update(soap, url, version);
    else if (do_send_inventory == 1)
        send_inventory(soap, url, xml);
    else if (do_send_net_usage == 1)
        send_net_usage(soap, url, xml);

    soap_end(soap);
    soap_free(soap);

    exit(0);
}

/*----------------------------------------------------------------------------
 * Summary: Contact webservice server to receive the update link.
 * Parameters:
 *      soap            SOAP struct
 *      url             URL of the contacted webservice server
 * Return: void
 */
void print_update_link(struct soap *soap, char *url)
{
    struct _ns1__getUpdateLink data;
    struct _ns1__getUpdateLinkResponse response;
    int error;
    int os = OS;
    data.args0 = &os;
    error = soap_call___ns2__getUpdateLink(soap, url, NULL, &data, &response);
    char *updateLink = response.return_;

    if (error != SOAP_OK)
    {
        soap_print_fault(soap, stderr);
        exit(error == NIC_ERROR ? error : 2);
    }

    /* Get update link from response */
    DEBUG("UpdateLink: %s\n", updateLink);

    /* Verify received update link and print appropriate messages */
    if (updateLink == NULL)
    {
        ERROR("ERROR (10): Get update link error (server returned NULL).\n");
        exit(10);
    }
    else if (strncmp(updateLink, "ERROR: ", 7) == 0)
    {
        ERROR("ERROR (20): (updateLink) Server return: %s\n", updateLink);
        exit(20);
    }

    /* Print to stdout the update link string */
    fprintf(stdout, "%s\n", updateLink);
}

/*----------------------------------------------------------------------------
 * Summary: Contact webservice server to verify client version.
 * Parameters:
 *      soap            SOAP struct
 *      url             URL of the contacted webservice server
 *      version         Current agent version
 * Return: void
 */
void update(struct soap *soap, char *url, char *version)
{
    struct _ns1__getAgentVersion data;
    struct _ns1__getAgentVersionResponse response;
    int error;
    int os = OS;
    data.args0 = &os;
    error = soap_call___ns2__getAgentVersion(soap, url, NULL, &data, &response);
    char *received_version = response.return_;

    if (error != SOAP_OK)
    {
        soap_print_fault(soap, stderr);
        exit(error == NIC_ERROR ? error : 2);
    }

    DEBUG("Version from server: %s\n", received_version);

    /* Check received version string for errors */
    if (received_version == NULL)
    {
        ERROR("ERROR (7): Server returned an empty version string.\n");
        exit(7);
    }
    else if (strncmp(received_version, "ERROR: ", 7) == 0)
    {
        ERROR("ERROR (8): Server returned: %s\n", received_version);
        exit(8);
    }
    /* Compare client and server versions */
    else if (strcmp(version, received_version) != 0)
    {
        ERROR("ERROR (4): Client version out-of-date. Update client.\n");
        print_update_link(soap, url);
        exit(4);
    }

    /* Client version up-to-date */
    fprintf(stderr, "Client version up-to-date\n");
}

/*----------------------------------------------------------------------------
 * Summary: Read XML data file and return the xmlInventory string.
 * Parameters:
 *      xmlpath     String with xml path
 * Return: String with xml data.
 */
char *read_xml_file(char *xmlpath)
{
    FILE *xmlFile;
    int aux;
    int i = 0;
    int tam = 1;
    char *xmlString = NULL;

    /* Open XML collect data file and check for errors */
    xmlFile = fopen(xmlpath, "r");
    if (xmlFile == NULL)
    {
        ERROR("ERROR (12): Problem while trying to open the xml file");
        exit(12);
    }

    /* Read and copy from data file to string */
    xmlString = (char *) malloc(sizeof(char));
    while ((aux = fgetc(xmlFile)) != EOF)
    {
        if (tam <= i)
        {
            tam *= 2;
            xmlString = (char *) realloc(xmlString, tam * sizeof(char));
        }
        if (xmlString == NULL)
        {
            ERROR("ERROR (12): Memory allocation error for xmlString.\n");
            exit(12);
        }
        xmlString[i++] = aux;
    }

    fclose(xmlFile);
    return xmlString;
}

/*----------------------------------------------------------------------------
 * Summary: Contact webservice server, sending the collected XML data.
 * Parameters:
 *      soap        SOAP struct
 *      url         URL of the contacted webservice server
 *      xml         String with the xml path
 * Return: void
 */
void send_inventory(struct soap *soap, char *url, char *xml)
{
    struct _ns1__setInventory data;
    struct _ns1__setInventoryResponse response;
    char *invResult;
    int error;

    data.args0 = read_xml_file(xml);

    error = soap_call___ns2__setInventory(soap, url, NULL, &data, &response);
    if (error != SOAP_OK)
    {
        soap_print_fault(soap, stderr);
        exit(error == NIC_ERROR ? error : 2);
    }
    invResult = response.return_;

    /* Get inventory result from response */
    DEBUG("InvResult: %s\n", invResult);

    /* Verify response and print appropriate message */
    if (strcmp(invResult, "Success") == 0)
    {
        fprintf(stdout, "XML data sent successfully\n");
    }
    else
    {
        ERROR("ERROR (5): %s.\n", invResult);
        exit(5);
    }
}

/*----------------------------------------------------------------------------
 * Summary: Contact webservice server, sending the collected XML data.
 * Parameters:
 *      soap        SOAP struct
 *      url         URL of the contacted webservice server
 *      xml         String with the xml path
 * Return: void
 */
void send_net_usage(struct soap *soap, char *url, char *xml)
{
    struct _ns1__setNetUsage data;
    struct _ns1__setNetUsageResponse response;
    char *invResult;
    int error;

    data.args0 = read_xml_file(xml);

    error = soap_call___ns2__setNetUsage(soap, url, NULL, &data, &response);
    if (error != SOAP_OK)
    {
        soap_print_fault(soap, stderr);
        exit(error == NIC_ERROR ? error : 2);
    }
    invResult = response.return_;

    /* Get inventory result from response */
    DEBUG("InvResult: %s\n", invResult);

    /* Verify response and print appropriate message */
    if (strcmp(invResult, "Success") == 0)
    {
        fprintf(stdout, "XML data sent successfully\n");
    }
    else
    {
        ERROR("ERROR (5): %s.\n", invResult);
        exit(5);
    }
}
